#!/usr/bin/env python
"""Checks that interactions between unlike types are correctly included also
when cutoff is longer than the like-like cutoffs"""

import unittest
import os
import numpy
import sys
import tempfile

sys.path.insert(1, "../../../lib")
sys.path.insert(1, "../../../Python")

import rumd
import rumd.Simulation
# it is essential that there is no doubt about which version is actually run!
print "File used to import the rumd module", rumd.__file__

pos = numpy.array([[0.000, 0.000, 0.000],
                   [1.200, 0.000, 0.000],
                   [0.000, 1.400, 0.000],
                   [0.000, 0.000, 1.600]])
N = len(pos)

xyz = "%d\nioformat=2 " % N
xyz += "sim_box=RectangularSimulationBox,10.00,10.00,10.00 numTypes=1  "
xyz += "mass=1.000 columns=type,x,y,z\n"

for pdx in xrange(N):
    this_pos = pos[pdx]
    xyz += "%d %.3f %.3f %.3f\n" % (0,
                                    this_pos[0],
                                    this_pos[1],
                                    this_pos[2])
    

def vLJ(sigma, epsilon, r):
    """ Evaluate the Lennard-Jones potential (no cutoff)"""
    return 4.0*epsilon*(pow(r/sigma, -12.0) - pow(r/sigma, -6.0))

class TestMultiplePotentials(unittest.TestCase):
    """Check that the exclusion list works as it should """
    def setUp(self):
        self.tdir = tempfile.mkdtemp()
        self.finp = self.tdir + '/test_exclusion_list.xyz'
        fh = open(self.finp, 'w')
        fh.write(xyz)
        fh.close()

    def test_multiple_potentials(self):
        sim = rumd.Simulation.Simulation(self.finp, verbose=False)
        # Create LJ pair-potential
        potential1 = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedForce)
        # Set up parameters for Lennard-Jones interactions
        # i  j   Sig   Eps  Rcut (in units of Sigma_ij)
        potential1.SetParams(0, 0, 1.00, 1.00, 5.0)

        potential2 = rumd.Pot_IPL_n(1, cutoff_method=rumd.ShiftedForce)
        potential2.SetParams(0, 0, 1.0, 1.0, 5.0)

        sim.SetPotential(potential1)
        sim.sample.CalcF()
        pe1 = sim.sample.GetPotentialEnergy()
        vir1 = sim.sample.GetVirial()
        forces1 = sim.sample.GetForces()
        
        sim.SetPotential(potential2)
        sim.sample.CalcF()
        pe2 = sim.sample.GetPotentialEnergy()
        vir2 = sim.sample.GetVirial()
        forces2 = sim.sample.GetForces()
        
        sim.AddPotential(potential1)
        sim.sample.CalcF()
        pe12 = sim.sample.GetPotentialEnergy()
        vir12 = sim.sample.GetVirial()
        forces12 = sim.sample.GetForces()
        
        #print pe1, pe2, pe12
        #print vir1, vir2, vir12
        #print forces1, forces2, forces12

        self.assertLess( abs(pe12 - (pe1+pe2) ), 1.e-6)

        self.assertLess( abs(vir12 - (vir1+vir2) ), 1.e-6)

        max_force_diff =  max(abs((forces12 - (forces1 + forces2) ).flatten() ))
        self.assertLess(max_force_diff, 1.e-6)

    def tearDown(self):
        import shutil
        if os.path.exists(self.finp):
            os.remove(self.finp)
        if os.path.exists(self.tdir):
            shutil.rmtree(self.tdir)


if __name__ == '__main__':
    unittest.main()
